/*
 * Copyright (C) 1995  Sam Moore, Warren Jasper
 * Copyright (C) 1997,1998,1999  Tim Edwards
 * All rights reserved.
 *
 * This software may be freely copied, modified, and redistributed
 * provided that this copyright notice is preserved on all copies.
 *
 * You may not distribute this software, in whole or in part, as part of
 * any commercial product without the express consent of the authors.
 *
 * There is no warranty or other guarantee of fitness of this software
 * for any purpose.  It is provided solely "as is".
 *
 */

/*----------------------------------------------------------------*/
/* ciodio96.c							  */
/*----------------------------------------------------------------*/

#include <linux/autoconf.h>
#include <linux/version.h>
#include <linux/module.h>
#include <linux/types.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/major.h>
#include <linux/mm.h>
#include <linux/timer.h>
#include <linux/kernel.h>
#include <linux/malloc.h>
#include <linux/string.h>
#include <linux/signal.h>
#include <linux/ioport.h>
#include <asm/io.h>
#include <asm/segment.h>
#include <asm/system.h>
#include <asm/uaccess.h>
#include "dio96.h"

/* For compatibility with Linux 2.1.x and backward compatibility */
/* with Linux 2.0.x. . .					 */

#ifdef get_fs_long
#define get_long(x,addr) (x) = get_fs_long(addr)
#else
#define get_long(x,addr) get_user((x), (const unsigned int *)(addr))
#endif

#ifndef put_fs_long
#define put_fs_long(x,addr) put_user((x),(unsigned int *)(addr))
#endif

/***************************************************************************
 *
 * Prototype of pubic and private functions.
 *
 ***************************************************************************/

#define BYTE unsigned char
#define LONG unsigned long

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

static int dio_read(struct inode *iNode, struct file *filePtr, char *buf, int count);
static int dio_write(struct inode *iNode, struct file *filePtr, const char *buf, int count);

#else

static ssize_t dio_read(struct file *filePtr, char *buf, size_t count, loff_t *ppos);
static ssize_t dio_write(struct file *filePtr, const char *buf, size_t count, loff_t *ppos);

#endif

static int dio_open(struct inode *iNode, struct file *filePtr);
static int dio_close(struct inode *iNode, struct file *filePtr);
static int dio_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd, LONG arg);

int init_module(void);
void cleanup_module(void) ;

/***************************************************************************
 *
 * Global data.
 *
 ***************************************************************************/

/* Default device major number */

#ifndef DEFAULT_MAJOR_DEV
#define DEFAULT_MAJOR_DEV  31    /* Default Major Device Number */
#endif

/* Base I/O Address */

#ifndef BASE_REG
#define BASE_REG         0x300   /* Default base address for NICB-DIO96 */
#endif

/* Number of channels */

#ifndef CHANNELS
#define CHANNELS	6
#endif

#define A_OFFSET	0
#define B_OFFSET	1
#define C_OFFSET	2
#define CTRL_OFFSET     3
#define PORT_OFFSET	4

#define A_MASK		0xff
#define B_MASK		0xff00
#define C_MASK		0xff0000

typedef struct CHANNEL_REC {
   int           open;        /* for lockout of other processes */
   unsigned char mode;	      /* for readback of 82C55 mode */
   unsigned char bits[3];     /* saved state of output */
} ChanRec;

static ChanRec Chan[CHANNELS];              /* Channel specific information   */
static int MajorNumber = DEFAULT_MAJOR_DEV; /* Major number compiled in       */

#ifndef TRUE
#define TRUE  1
#endif
#ifndef FALSE
#define FALSE 0
#endif

/***************************************************************************
 *
 * read() service function
 *
 ***************************************************************************/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

static int dio_read(struct inode *iNode, struct file *filePtr, char *buf, int count) 
{
  int   minor = MINOR(iNode->i_rdev);

#else

static ssize_t dio_read(struct file *filePtr, char *buf, size_t count, loff_t *ppos)
{
  int   minor = MINOR(filePtr->f_dentry->d_inode->i_rdev);

#endif

  long  value = 0L;
  int   portbase = BASE_REG + (minor * PORT_OFFSET);

  if (minor < CHANNELS) {
    value = inb(portbase + A_OFFSET);
    if (count > 1) value += inb(portbase + B_OFFSET) * 256;
    if (count > 2) value += inb(portbase + C_OFFSET) * 65536;

    cli();
    put_fs_long(value, buf);
    sti();
    return count;
  }

  printk("dio_read: Incorrect minor number (%d).\n", minor);
  return -1;
}

/***************************************************************************
 *
 * write() service function
 *
 ***************************************************************************/

#if LINUX_VERSION_CODE < KERNEL_VERSION(2,1,0)

static int dio_write(struct inode *iNode, struct file *filePtr, const char *buf,
        int count)
{
    int  minor = MINOR(iNode->i_rdev);

#else

static ssize_t dio_write(struct file *filePtr, const char *buf, size_t count,
	loff_t *ppos)
{
    int  minor = MINOR(filePtr->f_dentry->d_inode->i_rdev);

#endif
    int  portbase = BASE_REG + (minor * PORT_OFFSET);
    long value;
    BYTE bpart;

    if ( minor >= CHANNELS ) {
        printk("dio_write: bad minor number = %d\n", minor );
        return -1;
    }
    cli();
    get_long(value, buf);
    value &= 0xffffff;
    bpart = (BYTE)(value & A_MASK);
    Chan[minor].bits[0] = bpart;
    outb_p( bpart,        portbase + A_OFFSET);
    bpart = (BYTE)((value & B_MASK) >> 8);
    Chan[minor].bits[1] = bpart;
    outb_p( bpart,  portbase + B_OFFSET);
    bpart = (BYTE)((value & C_MASK) >> 16);
    Chan[minor].bits[2] = bpart;
    outb_p( bpart, portbase + C_OFFSET);
    sti();
    #ifdef DEBUG
        printk("DIO set to %#x\n", value & 0xf);
    #endif
    return 3;
}


/***************************************************************************
 *
 * open() service handler
 *
 ***************************************************************************/

static int dio_open(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);
    int portbase = BASE_REG + (minor * PORT_OFFSET);

    /* 
       check if device is already open: only one process may read from a
       port at a time.  There is still the possibility of two processes 
       reading from two different channels messing things up. However,
       the overhead to check for this may not be worth it.
    */

    if ( Chan[minor].open == TRUE ) {
        return -EBUSY;
    }

    Chan[minor].open = TRUE;              /* The device is open */
    Chan[minor].mode = 0;

#ifdef NICB
    if (minor == 5) {
       outb_p(0x0, portbase);
       outb_p(0x0, portbase + 1);
    }
    else
       outb_p(0x0, portbase + CTRL_OFFSET);
#else
    outb_p(0x0, portbase + CTRL_OFFSET);
#endif

    #ifdef DEBUG
        printk("%s: open(): minor %d mode %d.\n", ADAPTER_ID, minor, Chan[minor].mode);
    #endif

    MOD_INC_USE_COUNT;
    return (0);   
}

/***************************************************************************
 *
 * close() service handler
 *
 ***************************************************************************/

static int dio_close(struct inode *iNode, struct file *filePtr)
{
    int minor = MINOR(iNode->i_rdev);

    #ifdef DEBUG
        printk("%s: close() of minor number %d.\n", ADAPTER_ID, minor);
    #endif

    Chan[minor].open = FALSE;

    MOD_DEC_USE_COUNT;
    return (0);
}

/***************************************************************************
 *
 * iotctl() service handler
 *
 ***************************************************************************/

/* Note:
    Remember that FIOCLEX, FIONCLEX, FIONBIO, and FIOASYN are reserved ioctl cmd numbers
*/

/*------------------------------------------------------------------------------*/
/* IMPORTANT:  In this revision, ioctl() IGNORES the "minor" value and operates */
/* on the correct control register based on the ioctl command, provided that	*/
/* device has been opened (must be opened so it can be initialized. . .)	*/
/*										*/
/* ioctl() does NOT check to see if port has been opened already.		*/
/* ioctl() takes advantage of the 82C55's own bit-level commands for port C	*/
/*										*/
/* bit-wise and mode operations should not be allowed on the counter and	*/
/* interrupt registers of the 8253/8254, but for reasons of speed this is not	*/
/* checked.  Illegal registers can be written to, so it is up to the programmer */
/* to avoid the problem.							*/ 
/*------------------------------------------------------------------------------*/

static int dio_ioctl(struct inode *iNode, struct file *filePtr, unsigned int cmd,
	LONG arg)
{
  int minor = (cmd & 0x300) >> 8;
  int portbase = BASE_REG + (minor << 2);  /* PORT_OFFSET hard-coded */
  int channel = (cmd & 0x018) >> 3;

  /*--------------------------------------------------------------------*/
  /* ioctl command setup:						*/
  /*--------------------------------------------------------------------*/
  /* bits: 00pp smmc cbbb						*/
  /*--------------------------------------------------------------------*/
  /* p = port (0 = AP, 1 = BP, 2 = CP, 3 = DP)				*/
  /* s = set/get (0 = get, 1 = set)					*/
  /* m = mode (0 = clear (with s=1), 1 = mode, 2 = byte op, 3 = bit op	*/
  /* c = channel (0 = A, 1 = B, 2 = C)					*/
  /* b = bit (0 to 7)							*/
  /*--------------------------------------------------------------------*/

  cli();
  if (cmd & 0x080) {		/* SET operations */
     switch (cmd & 0x60) {
        case 0x60:			/* bit-wise set operation */
           Chan[minor].bits[channel] |= (0x01 << (cmd & 0x07));
	   if (channel == 2)
	      outb((BYTE)(0x01 | ((cmd & 0x07) << 1)), portbase + CTRL_OFFSET);
	   else
              outb((BYTE)(Chan[minor].bits[channel]), portbase + channel);
	   break;
        case 0x00:			/* bit-wise clear opearation */
           Chan[minor].bits[channel] &= ~(0x01 << (cmd & 0x07));
	   if (channel == 2)
	      outb((BYTE)((cmd & 0x07) << 1), portbase + CTRL_OFFSET);
	   else
              outb((BYTE)(Chan[minor].bits[channel]), portbase + channel);
	   break;
        case 0x40:			/* byte-wise operation */
           Chan[minor].bits[channel] = (BYTE)arg;
           outb((BYTE)arg, portbase + channel);
	   break;
        case 0x20:			/* mode-setting operation */
	   Chan[minor].mode = (BYTE)arg;
           outb_p((BYTE)(MODE_SET | arg), portbase + CTRL_OFFSET);
	   break;
     }
  }

  /*--------------------------------------------------------------------*/

  else {			/* GET operations */
     switch (cmd & 0x60) {
        case 0x60:			/* bit-wise operation */
	   put_fs_long(((inb(portbase + channel) & (0x01 << (cmd & 0x07))) > 0)
			? 1L : 0L, (long*)arg);
	   break;
        case 0x40:			/* byte-wise operation */
           put_fs_long(inb(portbase + channel), (long*)arg);
	   break;
        case 0x20:			/* mode-getting operation */
           put_fs_long(Chan[minor].mode, (long*) arg);
	   break;
        case 0x00:			/* no such command */
  	   sti();
           return(-EINVAL);
	   break;
     }
  }
  sti();

  /*--------------------------------------------------------------------*/

  return 0;
}

/***************************************************************************
 *
 * See /usr/include/linux/fs.h  for more information
 *
 ***************************************************************************/

struct file_operations dio_fops = {
	NULL,           /* lseek */
	dio_read,	/* read */
	dio_write,      /* write */
	NULL,		/* readdir */
	NULL,		/* select */
	dio_ioctl,	/* ioctl */
	NULL,		/* mmap */
	dio_open,	/* open */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	NULL,		/* flush */
#endif
	dio_close,	/* release */
	NULL,		/* fsync */
	NULL,		/* fasync */
	NULL,		/* check_media_change */
	NULL		/* revalidate */
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2,1,0)
	,NULL 		/* lock */
#endif
};

#ifdef __cplusplus
extern "C" {
#endif

/***************************************************************************
 *
 * Loads driver. Called when "insmod dio48.o" is invoked on the command line.
 *
 ***************************************************************************/

int init_module(void) 
{
  int i;

  /* Register as a device with kernel.  */

  if (register_chrdev(MajorNumber, "dio48", &dio_fops)) {
     printk("%s: Failure to load module.\n", ADAPTER_ID);
     return -EIO;
  }

  for (i=0; i < CHANNELS; i++)
     if (check_region(BASE_REG + i * PORT_OFFSET, 4) >= 0)
	request_region(BASE_REG + i * PORT_OFFSET, 4, "dio96");

  printk("%s: address=%#x.",ADAPTER_ID, BASE_REG);
  printk(" 1/18/99 tim.edwards@jhuapl.edu\n");

  cli();

  /* Set all channel structures to show nothing active/open */

  for (i = 0; i < CHANNELS; i++)
    Chan[i].open = FALSE;

  sti();
  return 0;
}

/***************************************************************************
 *
 * Remove driver. Called when "rmmod dio48" is run on the command line.
 *
 ***************************************************************************/

void cleanup_module(void) 
{
  int i;

  if (MOD_IN_USE) {
      printk("%s: device busy, remove delayed\n", ADAPTER_ID);
    }

    if (unregister_chrdev(MajorNumber, "dio48") != 0) {
      printk("%s: cleanup_module failed\n", ADAPTER_ID);
    } else {

    for (i=0; i < CHANNELS; i++)
       release_region(BASE_REG + i * PORT_OFFSET, 4);

    printk("%s: module removed.\n", ADAPTER_ID);

  }    
}

#ifdef __cplusplus
}
#endif

